
/*
	Example program for DVD-R access on Wii in GameCube mode.
    Copyright (C) 2008 WiiGator.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <malloc.h>
#include <ogcsys.h>
#include <gccore.h>
#include <unistd.h>

/* Error codes */
#define ERR_EINVAL		 -1
#define ERR_ENOENT		 -6
#define ERR_EIO			  2
#define ERR_EINCMD		128

/* Constants */
#define SECTOR_SIZE		0x800

syssram* __SYS_LockSram(void);
u32 __SYS_UnlockSram(u32 write);
u32 __SYS_SyncSram(void);

/* DVD drive */
volatile unsigned long *dvdio = (volatile unsigned long *) 0xCC006000;
volatile unsigned long *piio = (volatile unsigned long *) 0xCC003000;

unsigned char dvdbuf[SECTOR_SIZE] __attribute__ ((aligned(32)));

/* Variables */
unsigned long offset_base = 0;

static char hexmap[] = {
	'0', '1', '2', '3', '4', '5', '6', '7',
	'8', '9', 'a', 'b', 'c', 'd', 'e', 'f'
};

void debug_printf(const char *fmt, ...)
{
	char buf[1024];
	int len;
	va_list ap;

	usb_flush(1);
	va_start(ap, fmt);
	len = vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	fputs(buf, stdout);
#ifdef GEKKO_DEBUG
	usb_sendbuffer(GEKKO_CHANNEL, buf, len);
#endif
}


void hexDump(unsigned char *buffer, unsigned long size)
{
	int i;

	for (i = 0; i < size; i++) {
		if ((i > 0) && ((i & 0xF) == 0)) {
			debug_printf("\n");
		}
		unsigned char c;

		c = buffer[i];
		debug_printf("%c", hexmap[(c >> 4) & 0xf]);
		debug_printf("%c", hexmap[c & 0xf]);
		if ((i + 1) != size) {
			debug_printf(" ");
		}
	}
	debug_printf("\n");
}

unsigned long ack_cover_interrupt(void)
{
	unsigned int val;

	val = dvdio[0];
	debug_printf("DISR 0x%x\n", val);
	dvdio[0] = val;

	return val;
}

void dvd_reset(void)
{
	debug_printf("dvd_reset()\n");
	dvdio[1] = 2;
	unsigned long v = piio[9];

	piio[9] = (v & ~4) | 1;
	usleep(1000);
	piio[9] = v | 5;
}

int DI_Inquiry(void *outbuf)
{
	/* Reset notify */
	dvdio[0] = 0x2E;
	dvdio[1] = 0;
	dvdio[2] = 0x12000000;
	dvdio[3] = 0;
	dvdio[4] = 0x20;
	dvdio[5] = (unsigned long) outbuf;
	dvdio[6] = 0x20;
	dvdio[7] = 3;

	DCInvalidateRange(outbuf, 0x20);

	while (dvdio[7] & 1);

	/* Invalidate cache */
	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

unsigned int DI_GetError(void)
{
	/* Get last error. */
	dvdio[2] = 0xE0000000;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	return dvdio[8];
}

int DI_ReadDvd(void *outbuf, unsigned long len, unsigned long offset)
{
	unsigned long lba;

	lba = offset >> 11;
	/* Disc read */
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xD0000000;
	dvdio[3] = lba;
	dvdio[4] = (len + SECTOR_SIZE - 1) >> 11;
	dvdio[5] = (unsigned long) outbuf;
	dvdio[6] = len;
	dvdio[7] = 3;

	while (dvdio[7] & 1);

	/* Invalidate cache */
	DCInvalidateRange(outbuf, len);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

int DI_ReadDvdA8DiscId(void *outbuf)
{
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xA8000040;
	dvdio[3] = 0;
	dvdio[4] = 0x20;
	dvdio[5] = (unsigned long) outbuf;
	dvdio[6] = 0x20;
	dvdio[7] = 3;

	while (dvdio[7] & 1);

	/* Invalidate cache */
	DCInvalidateRange(outbuf, 0x20);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

int DI_ReadDvdA8(void *outbuf, unsigned long len, unsigned long offset)
{
	/* Disc read */
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xA8000000;
	dvdio[3] = offset >> 2;
	dvdio[4] = len;
	dvdio[5] = (unsigned long) outbuf;
	dvdio[6] = len;
	dvdio[7] = 3;

	while (dvdio[7] & 1);

	/* Invalidate cache */
	DCInvalidateRange(outbuf, len);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

int DI_StopMotor(unsigned long eject, unsigned long kill)
{
	/* Stop motor */
	dvdio[0] = 0x2E;
	dvdio[1] = 0;
	dvdio[2] = 0xE3000000;
	dvdio[3] = eject;
	dvdio[4] = kill;
	dvdio[5] = 0;
	dvdio[6] = 0;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

int DI_Offset(unsigned long offset)
{
	/* Set offset */
	dvdio[0] = 0x2E;
	dvdio[1] = 0;
	dvdio[2] = 0xD9000000;
	dvdio[3] = offset;
	dvdio[4] = 0;
	dvdio[5] = 0;
	dvdio[6] = 0;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

int DI_GetStatus(void *outbuf)
{
	/* Get cover state */
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xE0000000;
	dvdio[3] = 0;
	dvdio[4] = 0;
	dvdio[5] = 0;
	dvdio[6] = 0;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	/* Copy cover state */
	memcpy(outbuf, (void *) &dvdio[8], 4);

	return 1;
}

unsigned int DI_AudioBuffer(unsigned long subcmd)
{
	/* Get last error. */
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xE4000000 | subcmd;
	dvdio[3] = 0;
	dvdio[4] = 0;
	dvdio[5] = 0;
	dvdio[6] = 0;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

int DI_PlayAudio(unsigned long len, unsigned long offset, unsigned long subcmd)
{
	/* Disc read */
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xE1000000 | subcmd;
	dvdio[3] = offset >> 2;
	dvdio[4] = len;
	dvdio[5] = 0;
	dvdio[6] = 0;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

int DI_GetAudioStatus(unsigned long subcmd)
{
	/* Disc read */
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xE1000000 | subcmd;
	dvdio[3] = 0;
	dvdio[4] = 0;
	dvdio[5] = 0;
	dvdio[6] = 0;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

void time_test(void)
{
	time_t currentTime;
	syssram *sram;
	struct tm *gm;

	time(&currentTime);
	debug_printf("time 0x%08x\n", (unsigned long) currentTime);
	gm = gmtime(&currentTime);
	debug_printf("time: %s\n", asctime(gm));

	sram = __SYS_LockSram();
	debug_printf("GC SRAM Dump:\n");
	hexDump((unsigned char *) sram, 0x40);
	__SYS_UnlockSram(FALSE);
}

void wii_dvd_reset_unlock(void)
{
	volatile unsigned long *pi_cmd = (unsigned long *) 0xCC003024;
	volatile unsigned long *ios_magic = (unsigned long *) 0x800030F8;
	volatile unsigned long *ios_sync_base = (unsigned long *) 0x800030E0;
	volatile unsigned long *phys_ios_magic = (unsigned long *) 0xC00030F8;

	debug_printf("wii_dvd_reset_unlock()\n");

	/* I assume this will synchronize with Starlet/MIOS. */
	*pi_cmd |= 7;

	*ios_magic = 0xDEAEBEEF;
	DCFlushRange((void *) ios_sync_base, 32);

	while(*phys_ios_magic != 0) {
		/* I assume this waits for Starlet. */
	}
}

int dvdtest(void)
{
	unsigned long rv = -1;

	unsigned long offset;

	wii_dvd_reset_unlock();

	offset = 0;
	debug_printf("Gamecube DVD-R Test Program\n");
	debug_printf("Last error 0x%x\n", DI_GetError());
	if (!(dvdio[1] & 1)) {
		ack_cover_interrupt();
		dvd_reset();

		debug_printf("Read Disc ID\n");
		memset(dvdbuf, 0, SECTOR_SIZE);
		DCFlushRange(dvdbuf, SECTOR_SIZE);
		DI_ReadDvdA8DiscId(dvdbuf);
		debug_printf("Last error 0x%x\n", DI_GetError());
		hexDump(dvdbuf, 0x20);
		debug_printf("\n");

		debug_printf("Read DVD\n");
		memset(dvdbuf, 0, SECTOR_SIZE);
		DCFlushRange(dvdbuf, SECTOR_SIZE);
		rv = DI_ReadDvd(dvdbuf, SECTOR_SIZE, offset);
		debug_printf("rv = %d, Last error 0x%x\n", rv, DI_GetError());
		debug_printf("DI_ReadDvd offset 0x%x\n", offset);
		hexDump(dvdbuf, 0x40);

		debug_printf("Please eject disc\n");
		while (!(dvdio[1] & 1)) {
		}
		debug_printf("Please insert disc\n");
		while ((dvdio[1] & 1)) {
		}
		sleep(3);

		ack_cover_interrupt();
		dvd_reset();

		debug_printf("Read DVD\n");
		memset(dvdbuf, 0, SECTOR_SIZE);
		DCFlushRange(dvdbuf, SECTOR_SIZE);
		rv = DI_ReadDvd(dvdbuf, SECTOR_SIZE, offset);
		debug_printf("rv = %d, Last error 0x%x\n", rv, DI_GetError());
		debug_printf("DI_ReadDvd offset 0x%x\n", offset);
		hexDump(dvdbuf, 0x40);

		debug_printf("Enable Audio Buffer\n");
		rv = DI_AudioBuffer(0x10000);
		debug_printf("rv = %d, Last error 0x%x\n", rv, DI_GetError());

		debug_printf("Play Audio\n");
		rv = DI_PlayAudio(0x1000, 0, 0);
		debug_printf("rv = %d, Last error 0x%x\n", rv, DI_GetError());

		debug_printf("Get Audio Status 0x10000\n");
		rv = DI_GetAudioStatus(0x00010000);
		debug_printf("rv = %d, status 0x%x\n", rv, dvdio[8]);
		debug_printf("Last error 0x%x\n", DI_GetError());

		debug_printf("Get Audio Status 0x0\n");
		rv = DI_GetAudioStatus(0x0);
		debug_printf("rv = %d, status 0x%x\n", rv, dvdio[8]);
		debug_printf("Last error 0x%x\n", DI_GetError());

		debug_printf("Disable Audio Buffer\n", rv);
		rv = DI_AudioBuffer(0x00000);
		debug_printf("rv = %d, Last error 0x%x\n", rv, DI_GetError());

		debug_printf("Stop Motor\n");
		rv = DI_StopMotor(0, 0);
		debug_printf("rv = %d, Last error 0x%x\n", rv, DI_GetError());

	} else {
		debug_printf("Test failed no DVD inserted.\n");
	}
	time_test();
	return rv;
}
