/*
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
#include <stdio.h>
#include <unistd.h>
#include <ogcsys.h>
#include <gccore.h>
#include <ogc/cache.h>
#include <string.h>

#include "dvd.h"
#include "dvdreadpatch.h"
#include "dvdresetpatch.h"
#include "highplugin_bin.h"
#include "lowplugin_bin.h"
#include "gcplugin.h"
#include "debugprintf.h"
#include "loadmios.h"
#include "loadapp.h"
#include "systemconfig.h"

/** Time difference between UNIX time and GC time. */
#define GC_TIME_OFFSET 946684800
/** GameCube bus clock. */
#define GC_BUS_CLOCK 0x09A7EC80ULL
#define GC_BOOT_TIME ((unsigned long long *) 0x800030D8)

typedef struct {
	const char *gameid;
	u32 address1; /* Start address for patch search of second (fourth,...) DOL. */
	u32 size1; /* Size of patch. */
	u32 number1; /* Number of DOL changes. */
	u32 address2; /* Start address for patch search of first (third,...) DOL (Game Loader). */
	u32 size2; /* Size of Patch. */
	u32 number2; /* Number of DOL changes. */
} patch_entry_t;

/** Get timebase register, exported by libogc. */
extern long long gettime(void);

patch_entry_t game_patches[] = {
	{
		/* Zelda Collection. */
		.gameid = "PZLJ",

		/* Patch for Game Loader. */
		.address1 = 0x81307ebc,
		.size1 = 0x100,
		.number1 = 1,

		/* Patch for final Game. */
		.address2 = 0x8030a4ac,
		.size2 = 0x100,
		.number2 = 1,
	}
};

void nothing()
{
}

void setup_time(void)
{
	time_t currentTime;
	unsigned long long curTime;
	struct tm *gm;
	unsigned long long timebase;
	volatile unsigned long long *gcstarttime = GC_BOOT_TIME;

	time(&currentTime);
	debug_printf("time 0x%08x\n", (unsigned long) currentTime);
	gm = gmtime(&currentTime);
	debug_printf("time: %s\n", asctime(gm));

	/* Convert to GameCube time format. */
	curTime = currentTime;
	curTime -= GC_TIME_OFFSET;
	curTime = curTime * GC_BUS_CLOCK;
	timebase = gettime();
	timebase = curTime - timebase;

	*gcstarttime = timebase;
	DCFlushRange((void *) gcstarttime, 4);
	debug_printf("time 0x%08x%08x\n", (unsigned long) (*gcstarttime >> 32), (unsigned long) *gcstarttime);
}

void load_app(int reloader)
{
	gcplugin_jumptable_t *jmp1 = NULL;
	gcplugin_jumptable_t *jmp2 = NULL;
	u32 plugin_size = 0;
	const u8 *plugin = NULL;
	void *plugin_base = 0;
	int rv;
	u32 dvdid;
	static char gamecode[32] __attribute__((aligned(32)));

	void (*app_init)(void (*report)(const char* fmt, ...));
	int (*app_main)(void** dst, int* size, int* offset);
	void* (*app_final)();

	void (*app_entry)(void(**init)(void (*report)(const char* fmt, ...)),
		int (**main)(), void *(**final)());

	static char apploader_header[32] __attribute__((aligned(32)));

	setup_time();

	rv = dvd_read(gamecode, sizeof(gamecode), 0);
	DVD_CHECK();

	memcpy((void *) GAMECODE_ADDRESS, gamecode,  sizeof(gamecode));
	DCFlushRange((void *) GAMECODE_ADDRESS, sizeof(gamecode));

	dvdid = ((u32 *) gamecode)[0];

	rv = dvd_read(apploader_header, sizeof(apploader_header), APPLOADER_HEADER_OFFSET);
	DVD_CHECK();

	debug_printf("Reading apploader address 0x%x, len 0%x, offset 0x%x\n",
		(void*)APPLOADER_ADDRESS,
		((*(unsigned long*)(apploader_header + 0x14)) + 31) & ~31,
		APPLOADER_OFFSET);
	rv = dvd_read((void*)APPLOADER_ADDRESS,
		((*(unsigned long*)(apploader_header + 0x14)) + 31) & ~31,
		APPLOADER_OFFSET);
	DVD_CHECK();

	app_entry = (void (*)(void(**)(void (*)(const char*, ...)), int (**)(), void *(**)()))(*(unsigned long*)(apploader_header + 0x10));
	app_entry(&app_init, &app_main, &app_final);
	app_init((void (*)(const char*, ...))nothing);

	if (memcmp(gamecode, "GNHE", 4) == 0)  {
		plugin_size = highplugin_bin_size;
		plugin = highplugin_bin;
		plugin_base = HIGH_PLUGIN_BASE;
		setup_sys_config(-1);
	} else {
		plugin_size = lowplugin_bin_size;
		plugin = lowplugin_bin;
		plugin_base = LOW_PLUGIN_BASE;
		setup_sys_config(0);
	}
	jmp1 = (gcplugin_jumptable_t *) plugin_base;
	jmp2 = (gcplugin_jumptable_t *) (((u32) plugin_base) + sizeof(gcplugin_jumptable_t));

	/* Install DVD replacement functions. */
	memcpy(plugin_base, plugin, plugin_size);

	debug_printf("magic 0x%08x\n", jmp1->magic);

	/* Check for Reloader. */
	if (jmp1->magic == RELOAD_MAGIC) {
		int i;

		/* Reloader supported. */
		debug_printf("Reloader is supported.\n");

		if (reloader) {
			/* Try to patch every DVD transfer in the first DOL. */
			jmp2->reload_patch_start = 0;
			jmp2->reload_patch_size = 0;
			jmp2->reload_patch_number = 1;
		
			/* Try to patch every DVD transfer in the second DOL. */
			jmp1->reload_patch_start = 0;
			jmp1->reload_patch_size = 0;
			jmp1->reload_patch_number = 1;
		} else {
			for (i = 0; i < (sizeof(game_patches)/sizeof(patch_entry_t)); i++) {
				if (*((u32 *) game_patches[i].gameid) == dvdid) {
					debug_printf("Patch for game found.\n");
	
					/* Enable patching. */
					jmp2->reload_patch_start = game_patches[i].address1;
					jmp2->reload_patch_size = game_patches[i].size1;
					jmp2->reload_patch_number = game_patches[i].number1;
		
					jmp1->reload_patch_start = game_patches[i].address2;
					jmp1->reload_patch_size = game_patches[i].size2;
					jmp1->reload_patch_number = game_patches[i].number2;
				}
			}
		}
	} else {
		debug_printf("Reloader is not supported.\n");
	}

	DCFlushRange(plugin_base, plugin_size);
	ICInvalidateRange(plugin_base, plugin_size);

	while(1)
	{
		void* dst = 0;
		int len = 0,offset = 0;
		int res = app_main(&dst, &len, &offset);

		if (!res)
			break;
		debug_printf("Reading section buffer 0x%x, len 0%x, offset 0x%x\n", dst, len, offset);
		rv = dvd_read(dst, len, offset);
		DVD_CHECK();
		dvd_patchreset(dst, len, plugin_base);
		dvd_patchreadid(dst, len, plugin_base);
		dvd_patchread(dst, len, plugin_base, 0, gamecode);

		DCFlushRange(dst, len);
		ICInvalidateRange(dst, len);
	}
	DCFlushRange(plugin_base, plugin_size);
	ICInvalidateRange(plugin_base, plugin_size);

	debug_printf("Starting game by backuplauncher.\n");

	void (*entrypoint)() = app_final();
	entrypoint();
}

